@@ -16,44 +16,29 @@ from math import sqrt
 from time import sleep
 from pickle import PicklingError
 from multiprocessing import TimeoutError
+import pickle
 import pytest
 
+from importlib import reload
+
 import joblib
-from joblib import dump, load
 from joblib import parallel
+from joblib import dump, load
+from joblib.externals.loky import get_reusable_executor
 
 from joblib.test.common import np, with_numpy
 from joblib.test.common import with_multiprocessing
 from joblib.testing import (parametrize, raises, check_subprocess_call,
                             skipif, SkipTest, warns)
-from joblib._compat import PY3_OR_LATER, PY27
 
-try:
-    import cPickle as pickle
-except ImportError:
-    import pickle
 
-try:
-    from queue import Queue
-except ImportError:
-    # Backward compat
-    from Queue import Queue
+from queue import Queue
 
 try:
     import posix
 except ImportError:
     posix = None
 
-try:
-    RecursionError
-except NameError:
-    RecursionError = RuntimeError
-
-try:
-    reload         # Python 2
-except NameError:  # Python 3
-    from importlib import reload
-
 try:
     from ._openmp_test_helper.parallel_sum import parallel_sum
 except ImportError:
@@ -77,8 +62,6 @@ from joblib.parallel import effective_n_jobs, cpu_count
 
 from joblib.parallel import mp, BACKENDS, DEFAULT_BACKEND, EXTERNAL_BACKENDS
 from joblib.my_exceptions import JoblibException
-from joblib.my_exceptions import TransportableException
-from joblib.my_exceptions import JoblibValueError
 from joblib.my_exceptions import WorkerInterrupt
 
 
@@ -151,6 +134,23 @@ def test_effective_n_jobs():
     assert effective_n_jobs() > 0
 
 
+@pytest.mark.parametrize(
+    "backend_n_jobs, expected_n_jobs",
+    [(3, 3), (-1, effective_n_jobs(n_jobs=-1)), (None, 1)],
+    ids=["positive-int", "negative-int", "None"]
+)
+@with_multiprocessing
+def test_effective_n_jobs_None(backend_n_jobs, expected_n_jobs):
+    # check the number of effective jobs when `n_jobs=None`
+    # non-regression test for https://github.com/joblib/joblib/issues/984
+    with parallel_backend("threading", n_jobs=backend_n_jobs):
+        # when using a backend, the default of number jobs will be the one set
+        # in the backend
+        assert effective_n_jobs(n_jobs=None) == expected_n_jobs
+    # without any backend, None will default to a single job
+    assert effective_n_jobs(n_jobs=None) == 1
+
+
 ###############################################################################
 # Test parallel
 
@@ -234,6 +234,24 @@ def test_nested_parallel_warnings(parent_backend, child_backend, expected):
         assert all(res)
 
 
+@with_multiprocessing
+@parametrize('backend', ['loky', 'multiprocessing', 'threading'])
+def test_background_thread_parallelism(backend):
+    is_run_parallel = [False]
+
+    def background_thread(is_run_parallel):
+        with warns(None) as records:
+            Parallel(n_jobs=2)(
+                delayed(sleep)(.1) for _ in range(4))
+        print(len(records))
+        is_run_parallel[0] = len(records) == 0
+
+    t = threading.Thread(target=background_thread, args=(is_run_parallel,))
+    t.start()
+    t.join()
+    assert is_run_parallel[0]
+
+
 def nested_loop(backend):
     Parallel(n_jobs=2, backend=backend)(
         delayed(square)(.01) for _ in range(2))
@@ -399,11 +417,12 @@ def test_error_capture(backend):
 
     try:
         # JoblibException wrapping is disabled in sequential mode:
-        ex = JoblibException()
         Parallel(n_jobs=1)(
             delayed(division)(x, y) for x, y in zip((0, 1), (1, 0)))
     except Exception as ex:
         assert not isinstance(ex, JoblibException)
+    else:
+        raise ValueError("The excepted error has not been raised.")
 
 
 def consumer(queue, item):
@@ -514,16 +533,9 @@ def nested_function_outer(i):
 def test_nested_exception_dispatch(backend):
     """Ensure errors for nested joblib cases gets propagated
 
-    For Python 2.7, the TransportableException wrapping and unwrapping should
-    preserve the traceback information of the inner function calls.
-
-    For Python 3, we rely on the built-in __cause__ system that already
+    We rely on the Python 3 built-in __cause__ system that already
     report this kind of information to the user.
     """
-    if PY27 and backend == 'multiprocessing':
-        raise SkipTest("Nested parallel calls can deadlock with the python 2.7"
-                       "multiprocessing backend.")
-
     with raises(ValueError) as excinfo:
         Parallel(n_jobs=2, backend=backend)(
             delayed(nested_function_outer)(i) for i in range(30))
@@ -536,38 +548,7 @@ def test_nested_exception_dispatch(backend):
     assert 'nested_function_inner' in report
     assert 'exception_raiser' in report
 
-    if PY3_OR_LATER:
-        # Under Python 3, there is no need for exception wrapping as the
-        # exception raised in a worker process is transportable by default and
-        # preserves the necessary information via the `__cause__` attribute.
-        assert type(excinfo.value) is ValueError
-    else:
-        # The wrapping mechanism used to make exception of Python2.7
-        # transportable does not create a JoblibJoblibJoblibValueError
-        # despite the 3 nested parallel calls.
-        assert type(excinfo.value) is JoblibValueError
-
-
-def _reload_joblib():
-    # Retrieve the path of the parallel module in a robust way
-    joblib_path = Parallel.__module__.split(os.sep)
-    joblib_path = joblib_path[:1]
-    joblib_path.append('parallel.py')
-    joblib_path = '/'.join(joblib_path)
-    module = __import__(joblib_path)
-    # Reload the module. This should trigger a fail
-    reload(module)
-
-
-def test_multiple_spawning():
-    # Test that attempting to launch a new Python after spawned
-    # subprocesses will raise an error, to avoid infinite loops on
-    # systems that do not support fork
-    if not int(os.environ.get('JOBLIB_MULTIPROCESSING', 1)):
-        raise SkipTest()
-    with raises(ImportError):
-        Parallel(n_jobs=2, pre_dispatch='all')(
-            [delayed(_reload_joblib)() for i in range(10)])
+    assert type(excinfo.value) is ValueError
 
 
 class FakeParallelBackend(SequentialBackend):
@@ -731,7 +712,6 @@ def get_nested_pids():
                               for _ in range(2))
 
 
-
 class MyBackend(joblib._parallel_backends.LokyBackend):
     """Backend to test backward compatibility with older backends"""
     def get_nested_backend(self, ):
@@ -743,8 +723,8 @@ register_parallel_backend('back_compat_backend', MyBackend)
 
 
 @with_multiprocessing
-@pytest.mark.parametrize('backend', ['threading', 'loky', 'multiprocessing',
-                                     'back_compat_backend'])
+@parametrize('backend', ['threading', 'loky', 'multiprocessing',
+                         'back_compat_backend'])
 def test_nested_backend_context_manager(backend):
     # Check that by default, nested parallel calls will always use the
     # ThreadingBackend
@@ -759,8 +739,8 @@ def test_nested_backend_context_manager(backend):
 
 
 @with_multiprocessing
-@pytest.mark.parametrize('n_jobs', [2, -1, None])
-@pytest.mark.parametrize('backend', PARALLEL_BACKENDS)
+@parametrize('n_jobs', [2, -1, None])
+@parametrize('backend', PARALLEL_BACKENDS)
 def test_nested_backend_in_sequential(backend, n_jobs):
     # Check that by default, nested parallel calls will always use the
     # ThreadingBackend
@@ -787,6 +767,28 @@ def test_nested_backend_in_sequential(backend, n_jobs):
         )
 
 
+def check_nesting_level(inner_backend, expected_level):
+    with parallel_backend(inner_backend) as (backend, n_jobs):
+        assert backend.nesting_level == expected_level
+
+
+@with_multiprocessing
+@parametrize('outer_backend', PARALLEL_BACKENDS)
+@parametrize('inner_backend', PARALLEL_BACKENDS)
+def test_backend_nesting_level(outer_backend, inner_backend):
+    # Check that the nesting level for the backend is correctly set
+    check_nesting_level(outer_backend, 0)
+
+    Parallel(n_jobs=2, backend=outer_backend)(
+        delayed(check_nesting_level)(inner_backend, 1)
+        for _ in range(10)
+    )
+
+    with parallel_backend(inner_backend, n_jobs=2):
+        Parallel()(delayed(check_nesting_level)(inner_backend, 1)
+                   for _ in range(10))
+
+
 @with_multiprocessing
 def test_retrieval_context():
     import contextlib
@@ -825,16 +827,8 @@ def test_joblib_exception():
 
 def test_safe_function():
     safe_division = SafeFunction(division)
-    if PY3_OR_LATER:
-        with raises(ZeroDivisionError):
-            safe_division(1, 0)
-    else:
-        # Under Python 2.7, exception are wrapped with a special wrapper to
-        # preserve runtime information of the worker environment. Python 3 does
-        # not need this as it preserves the traceback information by default.
-        with raises(TransportableException) as excinfo:
-            safe_division(1, 0)
-        assert isinstance(excinfo.value.unwrap(), ZeroDivisionError)
+    with raises(ZeroDivisionError):
+        safe_division(1, 0)
 
     safe_interrupt = SafeFunction(interrupt_raiser)
     with raises(WorkerInterrupt):
@@ -872,23 +866,11 @@ def test_dispatch_race_condition(n_tasks, n_jobs, pre_dispatch, batch_size):
 
 @with_multiprocessing
 def test_default_mp_context():
+    mp_start_method = mp.get_start_method()
     p = Parallel(n_jobs=2, backend='multiprocessing')
     context = p._backend_args.get('context')
-    if sys.version_info >= (3, 4):
-        start_method = context.get_start_method()
-        # Under Python 3.4+ the multiprocessing context can be configured
-        # by an environment variable
-        env_method = os.environ.get('JOBLIB_START_METHOD', '').strip() or None
-        if env_method is None:
-            # Check the default behavior
-            if sys.platform == 'win32':
-                assert start_method == 'spawn'
-            else:
-                assert start_method == 'fork'
-        else:
-            assert start_method == env_method
-    else:
-        assert context is None
+    start_method = context.get_start_method()
+    assert start_method == mp_start_method
 
 
 @with_numpy
@@ -896,9 +878,6 @@ def test_default_mp_context():
 @parametrize('backend', PROCESS_BACKENDS)
 def test_no_blas_crash_or_freeze_with_subprocesses(backend):
     if backend == 'multiprocessing':
-        if sys.version_info < (3, 4):
-            raise SkipTest('multiprocessing can cause BLAS freeze on old '
-                           'Python that relies on fork.')
         # Use the spawn backend that is both robust and available on all
         # platforms
         backend = mp.get_context('spawn')
@@ -941,7 +920,7 @@ print(Parallel(n_jobs=2, backend=backend)(
 def test_parallel_with_interactively_defined_functions(backend):
     # When using the "-c" flag, interactive functions defined in __main__
     # should work with any backend.
-    if backend == "multiprocessing" and sys.platform == "win32":
+    if backend == "multiprocessing" and mp.get_start_method() != "fork":
         pytest.skip("Require fork start method to use interactively defined "
                     "functions with multiprocessing.")
     code = UNPICKLABLE_CALLABLE_SCRIPT_TEMPLATE_NO_MAIN.format(backend)
@@ -999,9 +978,7 @@ square = lambda x: x ** 2
 
 
 @with_multiprocessing
-@parametrize('backend', PROCESS_BACKENDS +
-             ([] if sys.version_info[:2] < (3, 4) or mp is None
-              else ['spawn']))
+@parametrize('backend', PROCESS_BACKENDS + ([] if mp is None else ['spawn']))
 @parametrize('define_func', [SQUARE_MAIN, SQUARE_LOCAL, SQUARE_LAMBDA])
 @parametrize('callable_position', ['delayed', 'args', 'kwargs'])
 def test_parallel_with_unpicklable_functions_in_args(
@@ -1182,12 +1159,6 @@ def test_warning_about_timeout_not_supported_by_backend():
 @parametrize('n_jobs', [1, 2, -2, -1])
 def test_abort_backend(n_jobs, backend):
     delays = ["a"] + [10] * 100
-
-    if os.environ.get("TRAVIS_OS_NAME") is not None and n_jobs < 0:
-        # Use only up to 8 cpu in travis as cpu_count return 32 whereas we
-        # only access 2 cores.
-        n_jobs += 8
-
     with raises(TypeError):
         t_start = time.time()
         Parallel(n_jobs=n_jobs, backend=backend)(
@@ -1215,16 +1186,26 @@ def test_memmapping_leaks(backend, tmpdir):
 
     # Make sure that the shared memory is cleaned at the end when we exit
     # the context
-    assert not os.listdir(tmpdir)
+    for _ in range(100):
+        if not os.listdir(tmpdir):
+            break
+        sleep(.1)
+    else:
+        raise AssertionError('temporary directory of Parallel was not removed')
 
     # Make sure that the shared memory is cleaned at the end of a call
     p = Parallel(n_jobs=2, max_nbytes=1, backend=backend)
     p(delayed(check_memmap)(a) for a in [np.random.random(10)] * 2)
 
-    assert not os.listdir(tmpdir)
+    for _ in range(100):
+        if not os.listdir(tmpdir):
+            break
+        sleep(.1)
+    else:
+        raise AssertionError('temporary directory of Parallel was not removed')
 
 
-@pytest.mark.parametrize('backend', [None, 'loky', 'threading'])
+@parametrize('backend', [None, 'loky', 'threading'])
 def test_lambda_expression(backend):
     # cloudpickle is used to pickle delayed callables
     results = Parallel(n_jobs=2, backend=backend)(
@@ -1233,9 +1214,6 @@ def test_lambda_expression(backend):
 
 
 def test_delayed_check_pickle_deprecated():
-    if sys.version_info < (3, 4):
-        pytest.skip("Warning check unstable under Python 2, life is too short")
-
     class UnpicklableCallable(object):
 
         def __call__(self, *args, **kwargs):
@@ -1497,7 +1475,8 @@ def test_thread_bomb_mitigation(backend):
 def _run_parallel_sum():
     env_vars = {}
     for var in ['OMP_NUM_THREADS', 'OPENBLAS_NUM_THREADS', 'MKL_NUM_THREADS',
-                'VECLIB_MAXIMUM_THREADS', 'NUMEXPR_NUM_THREADS']:
+                'VECLIB_MAXIMUM_THREADS', 'NUMEXPR_NUM_THREADS',
+                'NUMBA_NUM_THREADS', 'ENABLE_IPC']:
         env_vars[var] = os.environ.get(var)
     return env_vars, parallel_sum(100)
 
@@ -1505,12 +1484,18 @@ def _run_parallel_sum():
 @parametrize("backend", [None, 'loky'])
 @skipif(parallel_sum is None, reason="Need OpenMP helper compiled")
 def test_parallel_thread_limit(backend):
-    res = Parallel(n_jobs=2, backend=backend)(
+    results = Parallel(n_jobs=2, backend=backend)(
         delayed(_run_parallel_sum)() for _ in range(2)
     )
-    for value in res[0][0].values():
-        assert value == '1'
-    assert all([r[1] == 1 for r in res])
+    expected_num_threads = max(cpu_count() // 2, 1)
+    for worker_env_vars, omp_num_threads in results:
+        assert omp_num_threads == expected_num_threads
+        for name, value in worker_env_vars.items():
+            if name.endswith("_THREADS"):
+                assert value == str(expected_num_threads)
+            else:
+                assert name == "ENABLE_IPC"
+                assert value == "1"
 
 
 @skipif(distributed is not None,
@@ -1565,3 +1550,156 @@ def test_globals_update_at_each_parallel_call():
     workers_global_variable = Parallel(n_jobs=2)(
         delayed(check_globals)() for i in range(2))
     assert set(workers_global_variable) == {"changed value"}
+
+
+##############################################################################
+# Test environment variable in child env, in particular for limiting
+# the maximal number of threads in C-library threadpools.
+#
+
+def _check_numpy_threadpool_limits():
+    import numpy as np
+    # Let's call BLAS on a Matrix Matrix multiplication with dimensions large
+    # enough to ensure that the threadpool managed by the underlying BLAS
+    # implementation is actually used so as to force its initialization.
+    a = np.random.randn(100, 100)
+    np.dot(a, a)
+    from threadpoolctl import threadpool_info
+    return threadpool_info()
+
+
+def _parent_max_num_threads_for(child_module, parent_info):
+    for parent_module in parent_info:
+        if parent_module['filepath'] == child_module['filepath']:
+            return parent_module['num_threads']
+    raise ValueError("An unexpected module was loaded in child:\n{}"
+                     .format(child_module))
+
+
+def check_child_num_threads(workers_info, parent_info, num_threads):
+    # Check that the number of threads reported in workers_info is consistent
+    # with the expectation. We need to be carefull to handle the cases where
+    # the requested number of threads is below max_num_thread for the library.
+    for child_threadpool_info in workers_info:
+        for child_module in child_threadpool_info:
+            parent_max_num_threads = _parent_max_num_threads_for(
+                child_module, parent_info)
+            expected = {min(num_threads, parent_max_num_threads), num_threads}
+            assert child_module['num_threads'] in expected
+
+
+@with_numpy
+@with_multiprocessing
+@parametrize('n_jobs', [2, 4, -2, -1])
+def test_threadpool_limitation_in_child(n_jobs):
+    # Check that the protection against oversubscription in workers is working
+    # using threadpoolctl functionalities.
+
+    # Skip this test if numpy is not linked to a BLAS library
+    parent_info = _check_numpy_threadpool_limits()
+    if len(parent_info) == 0:
+        pytest.skip(msg="Need a version of numpy linked to BLAS")
+
+    workers_threadpool_infos = Parallel(n_jobs=n_jobs)(
+        delayed(_check_numpy_threadpool_limits)() for i in range(2))
+
+    n_jobs = effective_n_jobs(n_jobs)
+    expected_child_num_threads = max(cpu_count() // n_jobs, 1)
+
+    check_child_num_threads(workers_threadpool_infos, parent_info,
+                            expected_child_num_threads)
+
+
+@with_numpy
+@with_multiprocessing
+@parametrize('inner_max_num_threads', [1, 2, 4, None])
+@parametrize('n_jobs', [2, -1])
+def test_threadpool_limitation_in_child_context(n_jobs, inner_max_num_threads):
+    # Check that the protection against oversubscription in workers is working
+    # using threadpoolctl functionalities.
+
+    # Skip this test if numpy is not linked to a BLAS library
+    parent_info = _check_numpy_threadpool_limits()
+    if len(parent_info) == 0:
+        pytest.skip(msg="Need a version of numpy linked to BLAS")
+
+    with parallel_backend('loky', inner_max_num_threads=inner_max_num_threads):
+        workers_threadpool_infos = Parallel(n_jobs=n_jobs)(
+            delayed(_check_numpy_threadpool_limits)() for i in range(2))
+
+    n_jobs = effective_n_jobs(n_jobs)
+    if inner_max_num_threads is None:
+        expected_child_num_threads = max(cpu_count() // n_jobs, 1)
+    else:
+        expected_child_num_threads = inner_max_num_threads
+
+    check_child_num_threads(workers_threadpool_infos, parent_info,
+                            expected_child_num_threads)
+
+
+@with_multiprocessing
+@parametrize('n_jobs', [2, -1])
+@parametrize('var_name', ["OPENBLAS_NUM_THREADS",
+                          "MKL_NUM_THREADS",
+                          "OMP_NUM_THREADS"])
+def test_threadpool_limitation_in_child_override(n_jobs, var_name):
+    # Check that environment variables set by the user on the main process
+    # always have the priority.
+
+    # Clean up the existing executor because we change the environment of the
+    # parent at runtime and it is not detected in loky intentionally.
+    get_reusable_executor(reuse=True).shutdown()
+
+    def _get_env(var_name):
+        return os.environ.get(var_name)
+
+    original_var_value = os.environ.get(var_name)
+    try:
+        os.environ[var_name] = "4"
+        # Skip this test if numpy is not linked to a BLAS library
+        results = Parallel(n_jobs=n_jobs)(
+            delayed(_get_env)(var_name) for i in range(2))
+        assert results == ["4", "4"]
+
+        with parallel_backend('loky', inner_max_num_threads=1):
+            results = Parallel(n_jobs=n_jobs)(
+                delayed(_get_env)(var_name) for i in range(2))
+        assert results == ["1", "1"]
+
+    finally:
+        if original_var_value is None:
+            del os.environ[var_name]
+        else:
+            os.environ[var_name] = original_var_value
+
+
+@with_numpy
+@with_multiprocessing
+@parametrize('backend', ['multiprocessing', 'threading',
+                         MultiprocessingBackend(), ThreadingBackend()])
+def test_threadpool_limitation_in_child_context_error(backend):
+
+    with raises(AssertionError, match=r"does not acc.*inner_max_num_threads"):
+        parallel_backend(backend, inner_max_num_threads=1)
+
+
+@with_multiprocessing
+@parametrize('n_jobs', [2, 4, -1])
+def test_loky_reuse_workers(n_jobs):
+    # Non-regression test for issue #967 where the workers are not reused when
+    # calling multiple Parallel loops.
+
+    def parallel_call(n_jobs):
+        x = range(10)
+        Parallel(n_jobs=n_jobs)(delayed(sum)(x) for i in range(10))
+
+    # Run a parallel loop and get the workers used for computations
+    parallel_call(n_jobs)
+    first_executor = get_reusable_executor(reuse=True)
+
+    # Ensure that the workers are reused for the next calls, as the executor is
+    # not restarted.
+    for _ in range(10):
+        parallel_call(n_jobs)
+        executor = get_reusable_executor(reuse=True)
+        assert executor == first_executor
